package com.king011.fluter.mobile_audio_recorder

import android.app.Activity
import android.content.pm.PackageManager
import android.media.*
import android.media.MediaFormat.MIMETYPE_AUDIO_AAC
import android.os.Handler
import android.os.Looper
import android.os.Process
import android.util.Log
import androidx.annotation.NonNull
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.EventChannel
import io.flutter.plugin.common.EventChannel.EventSink
import io.flutter.plugin.common.PluginRegistry.Registrar
import io.flutter.plugin.common.PluginRegistry.RequestPermissionsResultListener
import java.nio.ByteBuffer
import java.nio.ByteOrder
import java.util.*
/** MobileAudioRecorderPlugin */
public class MobileAudioRecorderPlugin : FlutterPlugin, RequestPermissionsResultListener, EventChannel.StreamHandler, ActivityAware {

  // / Constants
  private val sampleRate = 44100
  private val bitRate = 96000
  private val logTag = "MobileAudioRecorderPlugin"

  // / Variables (i.e. will change value)
  private var channel: EventChannel? = null
  private var eventSink: EventSink? = null
  private var recording = false
  private var currentActivity: Activity? = null

  override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
    val messenger = flutterPluginBinding.getFlutterEngine().getDartExecutor()

    channel = EventChannel(messenger, "gitlab.com/king011/mobile_audio_recorder")
    channel!!.setStreamHandler(MobileAudioRecorderPlugin())
  }
  companion object {
    @JvmStatic
    fun registerWith(registrar: Registrar) {
        val channel = EventChannel(registrar.messenger(), "gitlab.com/king011/mobile_audio_recorder")
        channel.setStreamHandler(MobileAudioRecorderPlugin())
    }
  }

  override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
    channel!!.setStreamHandler(null)
  }

  override fun onDetachedFromActivity() {
      currentActivity = null
  }

  override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
      currentActivity = binding.activity
      binding.addRequestPermissionsResultListener(this)
  }

  override fun onAttachedToActivity(binding: ActivityPluginBinding) {
      currentActivity = binding.activity
      binding.addRequestPermissionsResultListener(this)
  }

  override fun onDetachedFromActivityForConfigChanges() {
      currentActivity = null
  }

  /**
   * Called from Flutter, starts the stream.
   */
  override fun onListen(arguments: Any?, events: EventSink?) {
      this.eventSink = events
      recording = true
      var format = "aac"
     if (arguments is String) {
        format = arguments
     }
      if (format == "pcm") {
          Log.i(logTag, "format PCM")
        streamPCMData()
      } else {
        Log.i(logTag, "format AAC")
        streamAACData()
      }
  }

  /**
   * Called from Flutter, which cancels the stream.
   */
  override fun onCancel(arguments: Any?) {
      recording = false
  }

  /**
   * Called by the plugin itself whenever it detects that permissions have not been granted.
   */
  override fun onRequestPermissionsResult(requestCode: Int, permissions: Array<String>, grantResults: IntArray): Boolean {
      val requestAudioPermissionCode = 200
      when (requestCode) {
          requestAudioPermissionCode -> if (grantResults[0] == PackageManager.PERMISSION_GRANTED) return true
      }
      return false
  }

  private fun streamPCMData() {
        Thread(Runnable {
            Process.setThreadPriority(Process.THREAD_PRIORITY_AUDIO)
            val bufferSize = AudioRecord.getMinBufferSize(sampleRate, AudioFormat.CHANNEL_IN_MONO, AudioFormat.ENCODING_PCM_16BIT)
            val audioBuffer = ShortArray(bufferSize / 2)
            val record = AudioRecord(
                    MediaRecorder.AudioSource.MIC,
                    sampleRate,
                    AudioFormat.CHANNEL_IN_MONO,
                    AudioFormat.ENCODING_PCM_16BIT,
                    bufferSize)
            if (record.state != AudioRecord.STATE_INITIALIZED) {
                Log.e(logTag, "Audio Record can't initialize!")
                return@Runnable
            }
            /** 開始錄音循環 */
            record.startRecording()
            while (recording) {
                val shortOut = record.read(audioBuffer, 0, audioBuffer.size)
                if (shortOut <1) {
                    continue
                }
                val byteBuffer = ByteBuffer.allocate(shortOut * 2)
                byteBuffer.order(ByteOrder.LITTLE_ENDIAN).asShortBuffer().put(audioBuffer)
                Handler(Looper.getMainLooper()).post {
                    eventSink!!.success(byteBuffer.array())
                }
            }
            record.stop()
            record.release()
        }).start()
    }
    private fun streamAACData() {
        Thread(Runnable {
            Process.setThreadPriority(Process.THREAD_PRIORITY_AUDIO)
            val bufferSize = AudioRecord.getMinBufferSize(sampleRate, AudioFormat.CHANNEL_IN_MONO, AudioFormat.ENCODING_PCM_16BIT)
            val audioBuffer = ShortArray(bufferSize / 2)
            val record = AudioRecord(
                    MediaRecorder.AudioSource.MIC,
                    sampleRate,
                    AudioFormat.CHANNEL_IN_MONO,
                    AudioFormat.ENCODING_PCM_16BIT,
                    bufferSize)
            if (record.state != AudioRecord.STATE_INITIALIZED) {
                Log.e(logTag, "Audio Record can't initialize!")
                return@Runnable
            }
            val mMediaCodec = createAACMediaCodec()

            /** 開始錄音循環 */
            mMediaCodec.start()
            val inputBuffers = mMediaCodec.getInputBuffers()
            val outputBuffers = mMediaCodec.getOutputBuffers()
            record.startRecording()
            while (recording) {
                val shortOut = record.read(audioBuffer, 0, audioBuffer.size)
                if (shortOut < 1) {
                    continue
                }
                val buffer = ByteBuffer.allocate(shortOut * 2)
                buffer.order(ByteOrder.LITTLE_ENDIAN).asShortBuffer().put(audioBuffer)

                val inputBufferIndex = mMediaCodec.dequeueInputBuffer(-1)
                if (inputBufferIndex >= 0) {
                    val inputBuffer = inputBuffers[inputBufferIndex]
                    inputBuffer.clear()
                    inputBuffer.put(buffer)
                    inputBuffer.limit(buffer.limit())
                    mMediaCodec.queueInputBuffer(inputBufferIndex, 0, buffer.limit(), 0, 0)
                }
                val bufferInfo = MediaCodec.BufferInfo()
                var outputBufferIndex = mMediaCodec.dequeueOutputBuffer(bufferInfo, 0)
                while (outputBufferIndex >= 0) {
                    val outputBuffer = outputBuffers[outputBufferIndex]
                     outputBuffer.position(bufferInfo.offset)
                    outputBuffer.limit(bufferInfo.offset + bufferInfo.size)
                     val chunkAudio = ByteArray(bufferInfo.size + 7) // 7 is ADTS size
                     addADTStoPacket(chunkAudio, chunkAudio.size)
                     outputBuffer.get(chunkAudio, 7, bufferInfo.size)
                     outputBuffer.position(bufferInfo.offset)
                     Handler(Looper.getMainLooper()).post {
                        eventSink!!.success(chunkAudio)
                    }
                    mMediaCodec.releaseOutputBuffer(outputBufferIndex, false)
                    outputBufferIndex = mMediaCodec.dequeueOutputBuffer(bufferInfo, 0)
                }
            }
            record.stop()
            record.release()
            mMediaCodec.stop()
            mMediaCodec.release()
        }).start()
    }
    private fun createAACMediaCodec(): MediaCodec {
        val format = MediaFormat.createAudioFormat(MIMETYPE_AUDIO_AAC, sampleRate, 1)
        format.setInteger(MediaFormat.KEY_AAC_PROFILE, MediaCodecInfo.CodecProfileLevel.AACObjectLC)
        format.setInteger(MediaFormat.KEY_BIT_RATE, bitRate)

        val mMediaCodec = MediaCodec.createEncoderByType(MIMETYPE_AUDIO_AAC)
         mMediaCodec.configure(format, null, null, MediaCodec.CONFIGURE_FLAG_ENCODE)
         return mMediaCodec
    }
    private fun addADTStoPacket(packet: ByteArray, packetLen: Int) {
        val profile = 2; // AAC LC
        val freqIdx = 4; // 44.1KHz
        val chanCfg = 1; // CPE
        // fill in ADTS data
        packet[0] = 0xFF.toByte()
        packet[1] = 0xF9.toByte()
        packet[2] = (((profile - 1) shl 6) + (freqIdx shl 2) + (chanCfg shr 2)).toByte()
        packet[3] = (((chanCfg and 3) shl 6) + (packetLen shr 11)).toByte()
        packet[4] = ((packetLen and 0x7FF) shr 3).toByte()
        packet[5] = (((packetLen and 7) shl 5) + 0x1F).toByte()
        packet[6] = 0xFC.toByte()
    }
}
