import 'dart:async';
import 'dart:typed_data';
import 'package:flutter/material.dart';
import 'package:permission_handler/permission_handler.dart';
import 'package:flutter/services.dart';

enum Status {
  Starting,
  Started,
  Stopping,
  Stoped,
}
const FormatAAC = 'aac';
const FormatPCM = 'pcm';

class MobileAudioRecorder {
  static Status _status = Status.Stoped;
  // static const MethodChannel _channel = const MethodChannel(
  //     'gitlab.com/king011/mobile_audio_recorder MethodChannel');

  static Future<String> get platformVersion async {
    // final String version = await _channel.invokeMethod('getPlatformVersion');
    // return version;
    return 'ok';
  }

  static List<String> get formats => const [FormatAAC, FormatPCM];
  static const EventChannel _eventChannel =
      EventChannel('gitlab.com/king011/mobile_audio_recorder');

  /// Verify that it was granted
  static Future<bool> checkPermission() async =>
      Permission.microphone.request().isGranted;

  /// Request the microphone permission
  static Future<void> requestPermission() async =>
      Permission.microphone.request();

  static Stream<Uint8List> _streamAAC;
  static Stream<Uint8List> _streamPCM;
  static StreamSubscription<Uint8List> _subscription;
  static Stream<Uint8List> _audioStream(String format) {
    Stream<Uint8List> stream;
    if (format == FormatAAC) {
      if (_streamAAC == null) {
        _streamAAC = _eventChannel
            .receiveBroadcastStream(format)
            .map((event) => Uint8List.fromList(event));
      }
      stream = _streamAAC;
    } else {
      if (_streamPCM == null) {
        _streamPCM = _eventChannel
            .receiveBroadcastStream(format)
            .map((event) => Uint8List.fromList(event));
      }
      stream = _streamPCM;
    }
    return stream;
  }

  static Future<void> start({
    String format = FormatAAC,
    @required Function(Uint8List) onData,
  }) async {
    assert(format != null);
    assert(onData != null);
    // 驗證狀態
    if (_status != Status.Stoped) {
      throw Exception('recorder not stopped');
    }
    _status = Status.Starting;
    try {
      await _start(
        format: format,
        onData: onData,
      );
      _status = Status.Started;
    } catch (e) {
      _status = Status.Stoped;
      throw e;
    }
  }

  static _start({
    String format,
    bool debug,
    @required Function onData,
  }) async {
    bool granted = await checkPermission();
    if (!granted) {
      // 請求權限後 重試
      await requestPermission();

      granted = await checkPermission();
    }
    if (!granted) {
      throw Exception('Permission Denied');
    }
    // 監聽數據
    _subscription = _audioStream(format).listen(onData);
  }

  static Future<void> stop() async {
    // 驗證狀態
    if (_status != Status.Started) {
      throw Exception('recorder not stopped');
    }
    _status = Status.Stopping;
    try {
      if (_subscription != null) {
        _subscription.cancel();
        _subscription = null;
      }
      _status = Status.Stoped;
    } catch (e) {
      _status = Status.Started;
      throw e;
    }
  }
}
